@@ -28,7 +28,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 
-#define VAL(x)  strtoul(x,(char **)NULL,0)
+#define VAL(x)  parsevaluestr(x)
 #define SECMAX  32
 #define SECNAME 16
 
@@ -62,19 +62,22 @@ int sparcleon0 = 0;
 int sparcleon0rom = 0;
     
 #ifndef TOOLBASE
-#define TOOLBASE "/opt/sparc-elf-3.2.3"
+#define TOOLBASE "/opt/mkprom2"
 #endif
 #ifndef RELEASE_VERSION
 #define RELEASE_VERSION "2.0.50"
 #endif
-//using -DTOOLBASE="\"...\""
-//#define TOOLBASE "/opt/sparc-elf"
+
+#if defined(__CYGWIN32__) || defined(__MINGW32__)
+    char *OS_EXESUFFIX = ".exe";
+#else
+    char *OS_EXESUFFIX = "";
+#endif
+
 const char version[] = "v" RELEASE_VERSION;
 int secnum = 0;
 FILE *dumpfile;
 
-int dobcholdcmd = 0;
-int dooldromcmd = 0;
 FILE *flashld = 0;
 char *flashldn = "dolinkromimg";
 char *flashsections[128];
@@ -84,6 +87,7 @@ int verbose = 0;
 int vverbose = 0;
 int leon = 1;
 double freq = 5E7;
+int romres = 0;
 int comp = 1;
 int flash = 0;
 int entry[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
@@ -95,10 +99,7 @@ char ofile[128] = "prom.out";
 
 enum {
     MULTIFLAG_FLAT  = 0x01,
-    MULTIFLAG_V8    = 0x02,
     MULTIFLAG_SOFT  = 0x04,
-    MULTIFLAG_UT699 = 0x08,
-    MULTIFLAG_B2BST = 0x10
 };
 /*
  * multiflags: Bit mask with MULTIFLAG_x values
@@ -112,13 +113,39 @@ char *prefix = TOOLBASE;
     
 void usage(char *);
 
-int searchfor(char *p) {
-    if (!system(p)) {
+int searchforcc(char *p) {
+    char cmd[256];
+#ifdef WIN32
+    snprintf(cmd, 256, "%s-gcc --version > NUL 2>&1", p);
+#else
+    snprintf(cmd, 256, "%s-gcc --version > /dev/null 2>&1", p);
+#endif
+    if (!system(cmd)) {
 	return 1;
     }
     return 0;
 }
 
+static unsigned int parsevaluestr(const char *valuestr)
+{
+    char *endptr;
+    unsigned int result;
+    char errmsg[256];
+
+    errno = 0;
+    result = strtoul(valuestr, &endptr, 0);
+
+    if (errno == 0 && *valuestr != 0 && *endptr == 0)
+        return result;
+
+    snprintf(errmsg, 256, "Unable to parse value \"%s\"", valuestr);
+    if (errno != 0)
+        perror(errmsg);
+    else
+        fprintf(stderr, "%s!\n", errmsg);
+    exit(1);
+}
+
 static inline void trysystem(const char *command)
 {
   int ret;
@@ -194,7 +221,7 @@ int main (argc, argv)
     int ddrbanks = 1;
     double refresh = 7.8;
     double ftmp;
-    int refr;
+    int refr = 390;
     char flist[512] = "";
     char xlist[512] = "";
     unsigned int mpstart = 0, mpirqsel0 = 0, mpirqsel1 = 0;
@@ -207,9 +234,6 @@ int main (argc, argv)
     unsigned int gptaddr = 0x80000300;
     unsigned int irqmpaddr = 0x80000200;
     int ecos = 0;
-    int dolinux = 0;
-    char *linux_cmdline = "console=ttyS0,38400", *linux_more = "";
-    unsigned int linux_base = 0x40000000;
     int mp = 0;
     int ncpu = 0;
     int uartnr = 0;
@@ -270,11 +294,7 @@ int main (argc, argv)
     char *lynxoshome = getenv( "ENV_PREFIX" );
     char lynxoshome_pre[1024];
 #endif
-#if defined(__CYGWIN32__) || defined(__MINGW32__)
-    char *postfix = ".exe";
-#else
-    char *postfix = "";
-#endif    
+    char *postfix = OS_EXESUFFIX;
     if (mkpromhome) {
 	prefix = mkpromhome;
     }
@@ -287,8 +307,8 @@ int main (argc, argv)
     prefix = lynxoshome_pre;
 #endif    
     
-    printf ("\nLEON2/3/ERC32 MKPROM prom builder for BCC, ECOS, RTEMS and ThreadX %s\n", version);
-    printf ("Copyright Cobham Gaisler AB 2004-2007, all rights reserved.\n\n");
+    printf ("MKPROM %s - boot image generator for LEON applications\n", version);
+    printf ("Copyright Cobham Gaisler AB 2004-2017, all rights reserved.\n\n");
 	if (argc < 2)
 	{
 		 usage(argv[0]);
@@ -396,14 +416,6 @@ int main (argc, argv)
 		  {
 		      dobch8 = 1;
 		  }
-                else if (strcmp (argv[stat], "-bcholdcmd") == 0)
-		  {
-		      dobcholdcmd = 1;
-		  }
-		else if (strcmp (argv[stat], "-oldromcmd") == 0)
-		  {
-		      dooldromcmd = 1;
-		  }
 		else if (strcmp (argv[stat], "-ddr2spa_cfg1") == 0)
 		  {
 		      ddr2spa_cfg1_cmd = 1; isddr = 1;
@@ -437,7 +449,10 @@ int main (argc, argv)
 			  }
 		  }
 
-		else if (strcmp (argv[stat], "-memcfg1") == 0)
+		else if (
+			strcmp(argv[stat], "-mcfg1") == 0 ||
+			strcmp(argv[stat], "-memcfg1") == 0
+		)
 		  {
 		      memcfg1_on = 1;
 		      if ((stat + 1) < argc)
@@ -446,7 +461,10 @@ int main (argc, argv)
 			  }
 		  }
 
-		else if (strcmp (argv[stat], "-memcfg2") == 0)
+		else if (
+			strcmp(argv[stat], "-mcfg2") == 0 ||
+			strcmp(argv[stat], "-memcfg2") == 0
+		)
 		  {
 		      memcfg2_on = 1;
 		      if ((stat + 1) < argc)
@@ -455,7 +473,10 @@ int main (argc, argv)
 			  }
 		  }
 
-		else if (strcmp (argv[stat], "-memcfg3") == 0)
+		else if (
+			strcmp(argv[stat], "-mcfg3") == 0 ||
+			strcmp(argv[stat], "-memcfg3") == 0
+		)
 		  {
 		      memcfg3_on = 1;
 		      if ((stat + 1) < argc)
@@ -463,7 +484,10 @@ int main (argc, argv)
 			      memcfg3 = (VAL (argv[++stat]));
 			  }
 		  }
-		else if (strcmp (argv[stat], "-sdmemcfg1") == 0)
+		else if (
+			strcmp(argv[stat], "-sdcfg1") == 0 ||
+			strcmp(argv[stat], "-sdmemcfg1") == 0
+		)
 		  {
 		      sdmemcfg1_on = 1;
 		      if ((stat + 1) < argc)
@@ -538,25 +562,6 @@ int main (argc, argv)
 		  {
 		      ecos = 1;
 		  }
-		else if (strcmp (argv[stat], "-linuxcmdline") == 0)
-                  {
-                    if ((stat + 1) < argc)
-                      linux_cmdline = argv[++stat];
-                  }
-		else if (strcmp (argv[stat], "-linuxmore") == 0)
-                  {
-                    if ((stat + 1) < argc)
-                      linux_more = argv[++stat];
-                  }
-		else if (strcmp (argv[stat], "-linuxbase") == 0)
-                  {
-                    if ((stat + 1) < argc)
-                      linux_base = VAL (argv[++stat]);
-                  }
-        	else if (strcmp (argv[stat], "-linux") == 0)
-		  {
-                      dolinux = 1;
-		  }
 		else if (strcmp (argv[stat], "-dsubaud") == 0)
 		  {
 		      if ((stat + 1) < argc)
@@ -940,6 +945,10 @@ int main (argc, argv)
 		  {
 		      checksvt = 1;
 		  }
+		else if (strcmp (argv[stat], "-romres") == 0)
+		  {
+		      romres = 1;
+		  }
 		else if (strcmp (argv[stat], "-qsvt") == 0) {
 		    qsvt = 1;
 		    goto copyxlist;
@@ -948,23 +957,11 @@ int main (argc, argv)
             multiflags |= MULTIFLAG_FLAT;
 		    goto copyxlist;
 		}
-		else if (strcmp (argv[stat], "-mv8") == 0 || strcmp (argv[stat], "-mcpu=v8") == 0) {
-            multiflags |= MULTIFLAG_V8;
-		    goto copyxlist;
-		}
 		else if (strcmp (argv[stat], "-msoft-float") == 0) {
             multiflags |= MULTIFLAG_SOFT;
 		    goto copyxlist;
 		}
-		else if (strcmp (argv[stat], "-mtune=ut699") == 0) {
-            multiflags |= MULTIFLAG_UT699;
-		    goto copyxlist;
-		}
-		else if (strcmp (argv[stat], "-mfix-b2bst") == 0) {
-            multiflags |= MULTIFLAG_B2BST;
-		    goto copyxlist;
-		}
-		
+
 		else
 		  {
 		  copyxlist:		      
@@ -975,26 +972,8 @@ int main (argc, argv)
 	    }
 	  else
 	    {
-                char *w = 0;
                 char *fn = argv[stat];
 
-                if (dolinux) {
-                    char c[1256];
-                    char *w = malloc(strlen(".prom.linux.wrap") + 16);
-                    strcpy(w, ".prom.linux.wrap");
-                    
-#ifdef __MINGW32__
-                    system("del .prom.linux.wrap");
-#else
-                    system ("rm -f .prom.linux.wrap");
-#endif
-                    sprintf(c,"sh %s/linuxwrap.sh %s %s -base 0x%x -freq 0x%x -cmdline \"%s\" %s ",prefix,fn,w, linux_base, (int)freq, linux_cmdline, linux_more);
-		    if (verbose)
-			    printf("Wrapping linux: %s\n",c);
-                    trysystem(c);
-                    fn = w;
-                }
-                
 		if (secnum == 0) {
 		    strcpy (filename, fn);
 		    strcat (flist, fn);
@@ -1017,12 +996,15 @@ int main (argc, argv)
 		}
 		*/
 
-                if (w)
-                    free(w);
 	    }
 	  stat++;
       }
 
+    if (strlen(flist) == 0) {
+        fprintf(stderr, "No input files specified!\n");
+        exit(1);
+    }
+
 	/* Setup defualt UART address */
 	if (!uartnr) {
 		if (leon2) {
@@ -1048,7 +1030,7 @@ int main (argc, argv)
 	}
     }
     
-    printf ("creating ");
+    printf ("Creating ");
     if (leon3) {
 	printf ("LEON3");
     }
@@ -1776,13 +1758,7 @@ int main (argc, argv)
       }
     fclose (dumpfile);
 
-    if(flash) {
-	if (ecos) {
-	    promcrt0 = "promcrt0_resident.o";
-	} else {
-	    promcrt0 = "promcrt0_resident_ecos.o";
-	}
-    } else {
+    if (!flash) {
 	if (!comp) {
 	    promload = "promload_copyonly.o";
 	} else {
@@ -1794,16 +1770,27 @@ int main (argc, argv)
     prefix = strlen(prefix) ? prefix : "." ;
     
     if(!ccprefixdo) {
-	printf("Searching for compiler to use (sparc-elf, sparc-rtems or sparc-linux):\n");
-	if (searchfor("sparc-elf-gcc --version")){
-	    ccprefix = "sparc-elf";
-	} else if (searchfor("sparc-gaisler-elf-gcc --version")){
-	    ccprefix = "sparc-gaisler-elf";
-	} else if (searchfor("sparc-rtems-gcc --version")){
-	    ccprefix = "sparc-rtems";
-	} else if (searchfor("sparc-linux-gcc --version")){
-	    ccprefix = "sparc-linux";
-	}
+        static char ccprefixes[5][20] = {"sparc-elf", "sparc-gaisler-elf",
+                                         "sparc-rtems", "sparc-rcc-rtems4.12",
+                                         "sparc-linux"};
+        int prefix_count = NELEM(ccprefixes);
+        int i;
+
+        for (i = 0; i < prefix_count; i++) {
+            if (searchforcc(ccprefixes[i])) {
+            ccprefix = ccprefixes[i];
+            break;
+            }
+        }
+
+        if ( i == prefix_count ) {
+            printf("No suitable compiler found!\n");
+            printf("Looked for the following compilers: \n");
+            for (i = 0; i < prefix_count; i++)
+                printf("\t%s-gcc\n", ccprefixes[i]);
+            printf("\n");
+            exit(1);
+        }
     }
 
     /* for sparcleon0 and execute-in-rom  modify the rom and ram fields of the linkerscript */
@@ -1842,11 +1829,7 @@ int main (argc, argv)
 	     
 	     ,
 	     ccprefix,
-#if defined(__CYGWIN32__) || defined(__MINGW32__)
-	     ".exe",
-#else
-	     "",
-#endif
+	     postfix,
 	     lscriptpath, startaddr
 	     
 	     );
@@ -2023,12 +2006,9 @@ int main (argc, argv)
     }
     if (flash) {
         /* change rom load addresses of <ofile> */
-	flash = 2;
-        if (dooldromcmd) {
-            /* old implementaion using --change-section-lma */
-            entry[0] = elf_load (ofile);
-        } else {
-            /* new implementaion using binary extraction */
+        flash = 2;
+        {
+            /* implementaion using binary extraction */
             char *b;
 #ifdef WIN32
             flashld = fopen (flashldn, "wb");
@@ -2047,7 +2027,10 @@ int main (argc, argv)
 #else
             sprintf(cmd, "cp %s %s.sym", ofile, ofile);
 #endif
-            if (verbose) printf("%s\n", cmd); fflush(stdout); system (cmd);
+            if (verbose)
+              printf("%s\n", cmd);
+            fflush(stdout);
+            system (cmd);
 
             /* traverse into the image, extract all sections and append linkerscript line so <flashldn> */
             entry[0] = elf_load (ofile);
@@ -2078,12 +2061,14 @@ int main (argc, argv)
 #else
                         sprintf(cmd,"rm -f %s\n",flashsections[i]);
 #endif
-                        if (verbose) printf("%s\n", cmd); fflush(stdout); system (cmd);
+                        if (verbose)
+                          printf("%s\n", cmd);
+                        fflush(stdout);
+                        system (cmd);
                     }
                 }
             }
         }
-        
     }
     if (dobch8) {
 	    appendbch8(".bch8",1, SEEK_END, startaddr + romsize);
@@ -2107,6 +2092,7 @@ int main (argc, argv)
 	    appendbch8(".bch8q", 0, SEEK_SET, romedacaddr);
     }
     
+    printf("Success!\n");
     exit (0);
 }
 
@@ -2122,16 +2108,18 @@ void usage (char *argv0)
 	 puts("  -freq <system_clock>\tDefines the system clock in MHz. This value is used to\n\t\t\tcalculate the divider value for the baud rate generator\n\t\t\tand the real-time clock. Default is 50 for LEON.");
 	 puts("  -noinit\t\tSuppress all code which initializes on-chip peripherals\n\t\t\tsuch as uarts, timers and memory controllers. This option\n\t\t\trequires -bdinit to add custom initialisation code,\n\t\t\tor the boot process will fail.");
 	 puts("  -nomsg\t\tSuppress the boot message.");
+	 puts("  -romres\t\tCreate ROM resident image.");
 	 puts("  -nocomp\t\tDon't compress application. Decreases loading time\n\t\t\ton the expense of rom size.");
 	 puts("  -o <outfile>\t\tPut the resulting image in outfile,\n\t\t\trather then prom.out (default).");
 	 puts("  -stack <addr>\t\tSets the initial stack pointer to addr.\n\t\t\tIf not specified, the stack starts at top-of-ram.");
 	 puts("  -v\t\t\tBe verbose; reports compression statistics\n\t\t\tand compile commands");
+	 puts("  -rstaddr <addr>\tSet the PROM start address. Default is 0x0.");
+
+	 puts("  -leon3\t\tGenerate an image for LEON3/4. This is the default.");
+	 puts("  -leon2\t\tGenerate an image for LEON2.");
+	 puts("  -agga4\t\tGenerate an image for AGGA4.");
+	 puts("  -erc32\t\tGenerate an image for ERC32.");
 
-	 puts("  -leon3 <addr>\t\tGenerate a image from leon3. (Default).");
-	 puts("  -leon2 <addr>\t\tGenerate a image from leon2. Default is Leon3.");
-	 puts("  -agga4 <addr>\t\tGenerate a image from agga4. Default is Leon3.");
-	 puts("  -erc32 <addr>\t\tGenerate a image from erc32. Default is Leon3.");
-	 
 	 puts("\nMkprom options for the LEON2 memory controller");
 	 puts("  -cas <delay>\t\tSet the SDRAM CAS delay. Allowed values are 2 and 3,\n\t\t\t2 is default.");
 	 puts("  -col <bits>\t\tSet the number of SDRAM column bits.\n\t\t\tAllowed values are 8 - 11, 9 is default.");
@@ -2151,9 +2139,9 @@ void usage (char *argv0)
 	 puts("  -trfc <delay>\t\tSet the SDRAM tRFC parameter (in ns). Default is 66 ns.");
 	 puts("  -trp <delay>\t\tSet the SDRAM tRP parameter (in ns). Default is 20 ns.");
 	 puts("  -refresh <delay>\tSet the SDRAM refresh period (in us). Default is 7.8 us,\n\t\t\talthough many SDRAMS actually use 15.6 us.");
-	 puts("  -memcfg1 <hex>\tSet memory configuration register 1. Overwrites calculated value");
-	 puts("  -memcfg2 <hex>\tSet memory configuration register 2. Overwrites calculated value");
-	 puts("  -memcfg3 <hex>\tSet memory configuration register 3. Overwrites calculated value");
+	 puts("  -mcfg1 <hex>\t\tSet memory configuration register 1. Overwrites calculated value");
+	 puts("  -mcfg2 <hex>\t\tSet memory configuration register 2. Overwrites calculated value");
+	 puts("  -mcfg3 <hex>\t\tSet memory configuration register 3. Overwrites calculated value");
 	 
 	 puts("\nMkprom options for the SPI memory controller");
 	 puts("  -spimeas\t\tEnable alternate scaler for SPI clock early in the boot process.");
@@ -2474,6 +2462,7 @@ elf_load (fname)
 {
     int tmp;
     FILE *xfile;
+    char errmsg[256];
 
 #ifdef WIN32
     xfile = fopen (fname, "rb");
@@ -2481,17 +2470,34 @@ elf_load (fname)
     xfile = fopen (fname, "r");
 #endif
 
-    if (xfile == NULL)
-      {
-	  printf ("open of %s failed\n", fname);
-	  return (-1);
-      }
-    if ((tmp = ldelf (xfile, dumpfile)) == -2) {
-	    if ((tmp = ldaout (xfile, dumpfile)) == -2) {
-	    }
+    if (xfile == NULL) {
+        snprintf(errmsg, 256, "Unable to open %s", fname);
+        perror(errmsg);
+        exit(1);
     }
-    
-    printf ("\n");
+
+    tmp = ldelf (xfile, dumpfile);
+
+    if (tmp == -1) {
+        fprintf(stderr, "%s is a not a valid ELF file!\n", fname);
+        exit(1);
+    }
+
+    if (tmp != -2)
+        return tmp;
+
+    tmp = ldaout (xfile, dumpfile);
+
+    if (tmp == -1) {
+        fprintf(stderr, "%s is a not a valid aout file!\n", fname);
+        exit(1);
+    }
+
+    if (tmp == -2) {
+        fprintf(stderr, "%s must be in SPARC elf32 or aout format!\n", fname);
+        exit(1);
+    }
+
     return (tmp);
 }
 
@@ -2501,49 +2507,23 @@ dumpsec (char *buf, int section_address, int section_size,
 {
     char cmd[512];
     char *lzss_buf;
-#if defined(__CYGWIN32__) || defined(__MINGW32__)
-    char *postfix = ".exe";
-#else
-    char *postfix = "";
-#endif
+    char *postfix = OS_EXESUFFIX;
 
     if ((secnum == 0) && (section_address == 0) && !(sparcleon0 && !sparcleon0rom))
       {
+        if (!romres) {
+            fprintf(stderr, "WARNING: Unexpected section address 0. Use -romres to create ROM resident image.\n");
+        }
 	printf("Section in rom detected, switching off compression\n");
 	  comp = 0;
 	  flash = 1;
-      }
+    } else if ((secnum == 0) && (section_address != 0)){
+        if (romres) {
+            fprintf(stderr, "WARNING: Option -romres expects section address 0.\n");
+        }
+    }
     if (flash) {
-	if (flash == 2) {
-            if (dooldromcmd) {
-	    if (strcmp(".text", section_name) == 0)
-	        foffset = section_size + section_address;
-	    else {
-#ifdef __MINGW32__
-	      /* Use copy since move complains the file is in use by other process */
-	      sprintf(cmd, "copy /y %s tmp.out", ofile);
-#else
-	      sprintf(cmd, "mv %s tmp.out", ofile);
-#endif
-	      if (verbose) printf("%s\n", cmd);
-	      fflush(stdout);
-	      system (cmd);
-              
-	      sprintf(cmd, "%s-objcopy%s --change-section-lma %s=0x%x tmp.out %s", 
-
-		      ccprefix,
-#if defined(__CYGWIN32__) || defined(__MINGW32__)
-	     ".exe",
-#else
-	     "",
-#endif
-		section_name, foffset + startaddr, ofile);
-	      if (verbose) printf("%s\n", cmd);
-	      fflush(stdout);
-	      trysystem (cmd);
-	      foffset += section_size;
-	    }
-            } else {
+        if (flash == 2) {
                 /* extract a section and convert to elf */
                 /* 1: extract section to binary <f>.b */
                 sprintf(cmd, "%s-objcopy%s --output-target=binary --only-section=%s %s %s%s.b",ccprefix,postfix,section_name, ofile, ofile, section_name);
@@ -2574,11 +2554,12 @@ dumpsec (char *buf, int section_address, int section_size,
 #else
                     sprintf(cmd,"rm -f %s%s.b\n",ofile, section_name);
 #endif
-                    if (verbose) printf("%s\n", cmd); fflush(stdout); system (cmd);
+                    if (verbose)
+                      printf("%s\n", cmd);
+                    fflush(stdout);
+                    system (cmd);
                 }
-                
-                
-                
+
                 /* assume that .text always come first */
                 if (strcmp(".text", section_name) == 0)
                     foffset = section_address;
@@ -2599,19 +2580,17 @@ dumpsec (char *buf, int section_address, int section_size,
                     flashsectionspos++;
                 }
                 foffset += section_size;
-            }
-            
-        
-	}
-	secnum++;
+        }
+        secnum++;
         return(0);
+    } else {
+        fprintf (dumpfile, "\t .text\n");
     }
-    else
-	fprintf (dumpfile, "\t .text\n");
     secarr[secnum].paddr = section_address;
     secarr[secnum].len = section_size;
     secarr[secnum].comp = comp;
-    strcpy (secarr[secnum].name, section_name);
+    strncpy (secarr[secnum].name, section_name, sizeof(secarr[secnum].name)-1);
+    secarr[secnum].name[sizeof(secarr[secnum].name)-1] = 0;
 
     fprintf (dumpfile, "\n\t.global _section%1d\n", secnum);
     fprintf (dumpfile, "_section%1d:\n", secnum);
@@ -2814,8 +2793,8 @@ ldelf (FILE * fp, FILE * dumpfile)
 		}
 	    }
 		    
-               if (verbose)
-		printf ("section: %s at 0x%x, size %d bytes\n",
+                if (verbose)
+                  printf ("section: %s at 0x%x, size %d bytes\n",
 			&strtab[sh.sh_name], sh.sh_addr, sh.sh_size);
 		mem = (char *) malloc (sh.sh_size);
 		if (mem != (char *) -1)
@@ -2823,7 +2802,8 @@ ldelf (FILE * fp, FILE * dumpfile)
 		      if (sh.sh_type == 1)
 			{
 			    fseek (fp, sh.sh_offset, SEEK_SET);
-			    fread (mem, sh.sh_size, 1, fp);
+			    if (fread (mem, sh.sh_size, 1, fp) != 1)
+				return -1;
 			    dumpsec (mem, sh.sh_addr, sh.sh_size,
 				     &strtab[sh.sh_name], dumpfile);
 			}
@@ -2938,11 +2918,7 @@ void appendbch8(char *post, int rev, int set, int pos )
 	char *bfn = "dump_bch8.b";
 	char *bfnw = "dump_bch8.b.w";
 	char ofilebch8[256];
-#if defined(__CYGWIN32__) || defined(__MINGW32__)
-		char *postfix = ".exe";
-#else
-		char *postfix = "";
-#endif
+	char *postfix = OS_EXESUFFIX;
 	
 	FILE *file, *bfile;
 
@@ -2970,7 +2946,10 @@ void appendbch8(char *post, int rev, int set, int pos )
 		memset(bchb, 0, blen + 8);
 		
 		fseek(file,0,SEEK_SET);
-		fread(buf, 1, off, file);
+		if (fread(buf, 1, off, file) != off) {
+			fprintf(stderr, "Unable to read %s!\n", ofn);
+			exit(1);
+		}
 		fclose (file);
 
 		if (rev) {
@@ -3012,8 +2991,6 @@ void appendbch8(char *post, int rev, int set, int pos )
 		
 		printf("Create bch8-containing %s from %s at 0x%x (romsize:0x%x)\n",ofilebch8,ofile,rompos, romsize);
 
-                if (!dobcholdcmd) {
-                    
                     /* the bch section, wrap into a elf image with section .bch */
                     sprintf (cmd, "%s-ld%s -r -b binary %s -o %s  ", ccprefix,postfix,bfn, bfnw);
                     bch8cmd(cmd);
@@ -3061,26 +3038,6 @@ void appendbch8(char *post, int rev, int set, int pos )
                         sprintf (cmd, "cp %s %s.renamed  ", ofilebch8, ofilebch8);
                         bch8cmd(cmd);
                     }
-                } else {
-                sprintf (cmd, "%s-objcopy%s --add-section .bch=%s %s %s  ", ccprefix,postfix,bfn, ofile, ofilebch8);
-		bch8cmd(cmd);
-                if (dodump) {
-                    sprintf (cmd, "cp %s %s.add.section  ", ofilebch8, ofilebch8);
-                    bch8cmd(cmd);
-                }
-		sprintf (cmd, "%s-objcopy%s  --set-section-flags .bch=alloc,load,data,contents %s  ", ccprefix,postfix,ofilebch8);
-		bch8cmd(cmd);
-                if (dodump) {
-                    sprintf (cmd, "cp %s %s.set.section.flags  ", ofilebch8, ofilebch8);
-                    bch8cmd(cmd);
-                }
-		sprintf (cmd, "%s-objcopy%s  --change-section-lma .bch=0x%x %s  ", ccprefix,postfix,rompos,ofilebch8);
-		bch8cmd(cmd);
-                if (dodump) {
-                    sprintf (cmd, "cp %s %s.change.section-lma  ", ofilebch8, ofilebch8);
-                    bch8cmd(cmd);
-                }
-                }
 
                 if (!dodump)
                     {
@@ -3142,20 +3099,10 @@ struct multientry {
 };
 
 static const struct multientry multientries[] = {
-    {0                                                                    , ""},
-    {                                                       MULTIFLAG_FLAT, "/flat"},
-    {                                       MULTIFLAG_V8                  , "/v8"},
-    {                                       MULTIFLAG_V8 |  MULTIFLAG_FLAT, "/v8/flat"},
-    {                   MULTIFLAG_SOFT                                    , "/soft"},
-    {                   MULTIFLAG_SOFT |                    MULTIFLAG_FLAT, "/soft/flat"},
-    {                   MULTIFLAG_SOFT |    MULTIFLAG_V8                  , "/soft/v8"},
-    {                   MULTIFLAG_SOFT |    MULTIFLAG_V8 |  MULTIFLAG_FLAT, "/soft/v8/flat"},
-    {MULTIFLAG_UT699                                                      , "/ut699"},
-    {MULTIFLAG_B2BST                                                      , "/mfix-b2bst"},
-    {MULTIFLAG_UT699 |                      MULTIFLAG_V8                  , "/v8/ut699"},
-    {MULTIFLAG_B2BST |                      MULTIFLAG_V8                  , "/v8/mfix-b2bst"},
-    {MULTIFLAG_B2BST |  MULTIFLAG_SOFT                                    , "/soft/mfix-b2bst"},
-    {MULTIFLAG_B2BST |  MULTIFLAG_SOFT |    MULTIFLAG_V8                  , "/soft/v8/mfix-b2bst"},
+    {0                                          , "/lib/ut699"},
+    {MULTIFLAG_SOFT                             , "/lib/ut699_soft"},
+    {                     MULTIFLAG_FLAT        , "/lib/ut699_flat"},
+    {MULTIFLAG_SOFT     | MULTIFLAG_FLAT        , "/lib/ut699_soft_flat"},
 };
 
 static const char *get_multidir(unsigned int multiflags)
